import * as _arktype_schema from '@arktype/schema';
import { Out, inferred, Root, inferNarrow, Predicate, Morph, BaseAttributes, inferMorphOut, NumericallyBoundable, writeUnboundableMessage, writeIndivisibleMessage, CheckResult, ProblemCode } from '@arktype/schema';
export { Out, Problem, Problems } from '@arktype/schema';
import { Dict, requireKeys, ErrorMessage, Completion, defined, extend, NumberLiteral, BigintLiteral, join, List, isAny, Constructor, conform, evaluate, intersectArrays, Hkt, Stringifiable, nominal, writeMalformedNumericLiteralMessage, objectKindOrDomainOf, isUnknown, requiredKeyOf, optionalKeyOf, equals, Primitive, CompiledFunction, Json, BuiltinObjects, BuiltinObjectKind, unionToTuple, entryOf, replaceKey, valueOf, Fn, isDisjoint, paramsOf, returnOf } from '@arktype/util';

interface InferredJsObjects {
    Function: Function;
    Date: Date;
    Error: Error;
    Map: Map<unknown, unknown>;
    RegExp: RegExp;
    Set: Set<unknown>;
    WeakMap: WeakMap<object, unknown>;
    WeakSet: WeakSet<object>;
    Promise: Promise<unknown>;
}
declare const jsObjects: RootScope<InferredJsObjects>;

interface InferredParsing {
    url: (In: string) => Out<URL>;
    number: (In: string) => Out<number>;
    integer: (In: string) => Out<number>;
    date: (In: string) => Out<Date>;
    json: (In: string) => Out<unknown>;
}
declare const parsingModule: Module<{
    exports: InferredParsing;
    locals: {};
    ambient: {};
}>;
type ParsingModule = typeof parsingModule;

interface InferredTsGenerics {
    Record: Generic<[
        "K",
        "V"
    ], {
        "[K]": "V";
    }, {}>;
}
declare const tsGenerics: RootScope<InferredTsGenerics>;

interface InferredTsKeywords {
    any: any;
    bigint: bigint;
    boolean: boolean;
    false: false;
    never: never;
    null: null;
    number: number;
    object: object;
    string: string;
    symbol: symbol;
    true: true;
    unknown: unknown;
    void: void;
    undefined: undefined;
}
declare const tsKeywords: RootScope<InferredTsKeywords>;

interface InferredValidation {
    alpha: string;
    alphanumeric: string;
    lowercase: string;
    uppercase: string;
    creditCard: string;
    email: string;
    uuid: string;
    url: string;
    semver: string;
    integer: number;
}
declare const validation: RootScope<InferredValidation>;

/** Root scopes can be inferred automatically from node definitions, but
 * explicitly typing them can improve responsiveness */
type RootScope<exports> = Scope<{
    exports: exports;
    locals: {};
    ambient: {};
}>;
type ArkResolutions = {
    exports: Ark;
    locals: {};
    ambient: Ark;
};
declare const ark: Scope<ArkResolutions>;
declare const arktypes: Module<ArkResolutions>;
interface Ark extends InferredTsKeywords, InferredJsObjects, InferredValidation, InferredTsGenerics {
    parse: ParsingModule;
}
declare const scope: ScopeParser<{}, Ark>;
declare const match: MatchParser<Ark>;
declare const when: WhenParser<Ark>;
declare const type: TypeParser<Ark>;
declare namespace type {
    type cast<to> = {
        [inferred]?: to;
    };
}
declare const define: DefinitionParser<Ark>;
declare const declare: DeclarationParser<Ark>;

type DateLiteral<source extends string = string> = `d"${source}"` | `d'${source}'`;

declare class Scanner<Lookahead extends string = string> {
    private chars;
    private i;
    constructor(def: string);
    /** Get lookahead and advance scanner by one */
    shift(): Lookahead;
    get lookahead(): Lookahead;
    get length(): number;
    shiftUntil(condition: Scanner.UntilCondition): string;
    shiftUntilNextTerminator(): string;
    shiftUntilNonWhitespace(): string;
    jumpToIndex(i: number): void;
    get location(): number;
    get unscanned(): string;
    get scanned(): string;
    sliceChars(start: number, end?: number): string;
    lookaheadIs<Char extends Lookahead>(char: Char): this is Scanner<Char>;
    lookaheadIsIn<Tokens extends Dict>(tokens: Tokens): this is Scanner<Extract<keyof Tokens, string>>;
}
declare namespace Scanner {
    type UntilCondition = (scanner: Scanner, shifted: string) => boolean;
    type OnInputEndFn = (scanner: Scanner, shifted: string) => string;
    type ShiftUntilOptions = {
        onInputEnd?: OnInputEndFn;
    };
    const lookaheadIsTerminator: UntilCondition;
    const lookaheadIsNotWhitespace: UntilCondition;
    const terminatingChars: {
        readonly "<": true;
        readonly ">": true;
        readonly "=": true;
        readonly "|": true;
        readonly "&": true;
        readonly ")": true;
        readonly "[": true;
        readonly "%": true;
        readonly " ": true;
        readonly ",": true;
    };
    type TerminatingChar = keyof typeof terminatingChars;
    const finalizingLookaheads: {
        readonly ">": true;
        readonly ",": true;
        readonly "": true;
    };
    type FinalizingLookahead = keyof typeof finalizingLookaheads;
    type InfixToken = Comparator | "|" | "&" | "%" | ":" | "=>";
    type PostfixToken = "[]";
    type OperatorToken = InfixToken | PostfixToken;
    const escapeToken = "\\";
    type EscapeToken = typeof escapeToken;
    const whiteSpaceTokens: {
        readonly " ": true;
        readonly "\n": true;
    };
    type WhiteSpaceToken = keyof typeof whiteSpaceTokens;
    const lookaheadIsFinalizing: (lookahead: string, unscanned: string) => lookahead is ">" | ",";
    type lookaheadIsFinalizing<lookahead extends string, unscanned extends string> = lookahead extends ">" ? unscanned extends `=${infer nextUnscanned}` ? nextUnscanned extends `=${string}` ? true : false : Scanner.skipWhitespace<unscanned> extends "" | `${TerminatingChar}${string}` ? true : false : lookahead extends "," ? true : false;
    type shift<lookahead extends string, unscanned extends string> = `${lookahead}${unscanned}`;
    type shiftUntil<unscanned extends string, terminator extends string, scanned extends string = ""> = unscanned extends shift<infer lookahead, infer nextUnscanned> ? lookahead extends terminator ? scanned extends `${infer base}${EscapeToken}` ? shiftUntil<nextUnscanned, terminator, `${base}${lookahead}`> : [scanned, unscanned] : shiftUntil<nextUnscanned, terminator, `${scanned}${lookahead}`> : [scanned, ""];
    type shiftUntilNot<unscanned extends string, nonTerminator extends string, scanned extends string = ""> = unscanned extends shift<infer lookahead, infer nextUnscanned> ? lookahead extends nonTerminator ? shiftUntilNot<nextUnscanned, nonTerminator, `${scanned}${lookahead}`> : [scanned, unscanned] : [scanned, ""];
    type shiftUntilNextTerminator<unscanned extends string> = shiftUntil<unscanned, TerminatingChar>;
    type skipWhitespace<unscanned extends string> = shiftUntilNot<unscanned, WhiteSpaceToken>[1];
    type shiftResult<scanned extends string, unscanned extends string> = [
        scanned,
        unscanned
    ];
}

type BranchState$1 = {
    prefixes: StringifiablePrefixOperator[];
    leftBound?: OpenLeftBound;
    "&"?: Root;
    "|"?: Root;
};
type DynamicStateWithRoot = requireKeys<DynamicState, "root">;
declare class DynamicState {
    readonly ctx: ParseContext;
    readonly scanner: Scanner;
    root: Root | undefined;
    branches: BranchState$1;
    finalizer: Scanner.FinalizingLookahead | undefined;
    groups: BranchState$1[];
    constructor(def: string, ctx: ParseContext);
    error(message: string): never;
    hasRoot(): this is DynamicStateWithRoot;
    unsetRoot(): this["root"];
    constrainRoot(...args: Parameters<Root["constrain"]>): void;
    setRoot(root: Root): void;
    finalize(finalizer: Scanner.FinalizingLookahead): undefined;
    reduceLeftBound(limit: LimitLiteral, comparator: Comparator): undefined;
    finalizeBranches(): void;
    finalizeGroup(): undefined;
    addPrefix(prefix: StringifiablePrefixOperator): void;
    applyPrefixes(): void;
    pushRootToBranch(token: "|" | "&"): void;
    private assertRangeUnset;
    reduceGroupOpen(): void;
    previousOperator(): "keyof" | ">" | ">=" | "|" | "&" | undefined;
    shiftedByOne(): this;
}

type StaticState = {
    root: unknown;
    branches: BranchState;
    groups: BranchState[];
    finalizer: Scanner.FinalizingLookahead | ErrorMessage | undefined;
    scanned: string;
    unscanned: string;
};
type BranchState = {
    prefixes: StringifiablePrefixOperator[];
    leftBound: OpenLeftBound | undefined;
    "&": unknown;
    "|": unknown;
};
declare namespace state {
    export type initialize<def extends string> = from<{
        root: undefined;
        branches: initialBranches;
        groups: [];
        finalizer: undefined;
        scanned: "";
        unscanned: def;
    }>;
    export type error<message extends string> = from<{
        root: undefined;
        branches: initialBranches;
        groups: [];
        finalizer: ErrorMessage<message>;
        scanned: "";
        unscanned: "";
    }>;
    export type completion<text extends string> = from<{
        root: undefined;
        branches: initialBranches;
        groups: [];
        finalizer: Completion<text>;
        scanned: "";
        unscanned: "";
    }>;
    type initialBranches = branchesFrom<{
        prefixes: [];
        leftBound: undefined;
        "&": undefined;
        "|": undefined;
    }>;
    type updateScanned<previousScanned extends string, previousUnscanned extends string, updatedUnscanned extends string> = previousUnscanned extends `${infer justScanned}${updatedUnscanned}` ? `${previousScanned}${justScanned}` : previousScanned;
    export type setRoot<s extends StaticState, root, unscanned extends string> = from<{
        root: root;
        branches: s["branches"];
        groups: s["groups"];
        finalizer: s["finalizer"];
        scanned: updateScanned<s["scanned"], s["unscanned"], unscanned>;
        unscanned: unscanned;
    }>;
    export type addPrefix<s extends StaticState, prefix extends StringifiablePrefixOperator, unscanned extends string> = from<{
        root: s["root"];
        branches: {
            prefixes: [...s["branches"]["prefixes"], prefix];
            leftBound: s["branches"]["leftBound"];
            "&": s["branches"]["&"];
            "|": s["branches"]["|"];
        };
        groups: s["groups"];
        finalizer: s["finalizer"];
        scanned: updateScanned<s["scanned"], s["unscanned"], unscanned>;
        unscanned: unscanned;
    }>;
    export type reduceBranch<s extends StaticState, token extends "|" | "&", unscanned extends string> = s["branches"]["leftBound"] extends {} ? openRangeError<s["branches"]["leftBound"]> : from<{
        root: undefined;
        branches: {
            prefixes: [];
            leftBound: undefined;
            "&": token extends "&" ? mergeToIntersection<s> : undefined;
            "|": token extends "|" ? mergeToUnion<s> : s["branches"]["|"];
        };
        groups: s["groups"];
        finalizer: s["finalizer"];
        scanned: updateScanned<s["scanned"], s["unscanned"], unscanned>;
        unscanned: unscanned;
    }>;
    export type reduceLeftBound<s extends StaticState, limit extends LimitLiteral, comparator extends Comparator, unscanned extends string> = comparator extends "<" | "<=" ? s["branches"]["leftBound"] extends {} ? state.error<writeMultipleLeftBoundsMessage<s["branches"]["leftBound"]["limit"], s["branches"]["leftBound"]["comparator"], limit, InvertedComparators[comparator]>> : from<{
        root: undefined;
        branches: {
            prefixes: s["branches"]["prefixes"];
            leftBound: {
                limit: limit;
                comparator: InvertedComparators[comparator];
            };
            "&": s["branches"]["&"];
            "|": s["branches"]["|"];
        };
        groups: s["groups"];
        finalizer: s["finalizer"];
        scanned: updateScanned<s["scanned"], s["unscanned"], unscanned>;
        unscanned: unscanned;
    }> : state.error<writeUnpairableComparatorMessage<comparator>>;
    export type reduceRange<s extends StaticState, minLimit extends LimitLiteral, minComparator extends MinComparator, maxComparator extends MaxComparator, maxLimit extends LimitLiteral, unscanned extends string> = state.from<{
        root: [minLimit, minComparator, [s["root"], maxComparator, maxLimit]];
        branches: {
            prefixes: s["branches"]["prefixes"];
            leftBound: undefined;
            "&": s["branches"]["&"];
            "|": s["branches"]["|"];
        };
        groups: s["groups"];
        finalizer: s["finalizer"];
        scanned: updateScanned<s["scanned"], s["unscanned"], unscanned>;
        unscanned: unscanned;
    }>;
    export type reduceSingleBound<s extends StaticState, comparator extends Comparator, limit extends LimitLiteral, unscanned extends string> = state.from<{
        root: [s["root"], comparator, limit];
        branches: {
            prefixes: s["branches"]["prefixes"];
            leftBound: undefined;
            "&": s["branches"]["&"];
            "|": s["branches"]["|"];
        };
        groups: s["groups"];
        finalizer: s["finalizer"];
        scanned: updateScanned<s["scanned"], s["unscanned"], unscanned>;
        unscanned: unscanned;
    }>;
    type mergeToUnion<s extends StaticState> = s["branches"]["|"] extends undefined ? mergeToIntersection<s> : [s["branches"]["|"], "|", mergeToIntersection<s>];
    type mergeToIntersection<s extends StaticState> = s["branches"]["&"] extends undefined ? mergePrefixes<s> : [s["branches"]["&"], "&", mergePrefixes<s>];
    type mergePrefixes<s extends StaticState, remaining extends unknown[] = s["branches"]["prefixes"]> = remaining extends [infer head, ...infer tail] ? [head, mergePrefixes<s, tail>] : s["root"];
    type popGroup<stack extends BranchState[], top extends BranchState> = [
        ...stack,
        top
    ];
    export type finalizeGroup<s extends StaticState, unscanned extends string> = s["branches"]["leftBound"] extends {} ? openRangeError<s["branches"]["leftBound"]> : s["groups"] extends popGroup<infer stack, infer top> ? from<{
        groups: stack;
        branches: top;
        root: mergeToUnion<s>;
        finalizer: s["finalizer"];
        scanned: updateScanned<s["scanned"], s["unscanned"], unscanned>;
        unscanned: unscanned;
    }> : state.error<writeUnmatchedGroupCloseMessage<unscanned>>;
    export type reduceGroupOpen<s extends StaticState, unscanned extends string> = from<{
        groups: [...s["groups"], s["branches"]];
        branches: initialBranches;
        root: undefined;
        finalizer: s["finalizer"];
        scanned: updateScanned<s["scanned"], s["unscanned"], unscanned>;
        unscanned: unscanned;
    }>;
    export type finalize<s extends StaticState, finalizer extends Scanner.FinalizingLookahead> = s["groups"] extends [] ? s["branches"]["leftBound"] extends {} ? openRangeError<s["branches"]["leftBound"]> : from<{
        root: mergeToUnion<s>;
        groups: s["groups"];
        branches: initialBranches;
        finalizer: finalizer;
        scanned: s["scanned"];
        unscanned: s["unscanned"];
    }> : state.error<writeUnclosedGroupMessage<")">>;
    type openRangeError<range extends defined<BranchState["leftBound"]>> = state.error<writeOpenRangeMessage<range["limit"], range["comparator"]>>;
    export type previousOperator<s extends StaticState> = s["branches"]["leftBound"] extends {} ? s["branches"]["leftBound"]["comparator"] : s["branches"]["prefixes"] extends [
        ...unknown[],
        infer tail extends string
    ] ? tail : s["branches"]["&"] extends {} ? "&" : s["branches"]["|"] extends {} ? "|" : undefined;
    export type scanTo<s extends StaticState, unscanned extends string> = from<{
        root: s["root"];
        branches: s["branches"];
        groups: s["groups"];
        finalizer: s["finalizer"];
        scanned: updateScanned<s["scanned"], s["unscanned"], unscanned>;
        unscanned: unscanned;
    }>;
    export type from<s extends StaticState> = s;
    export type branchesFrom<b extends BranchState> = b;
    export {};
}

type StringLiteral<Text extends string = string> = DoubleQuotedStringLiteral<Text> | SingleQuotedStringLiteral<Text>;
type DoubleQuotedStringLiteral<Text extends string = string> = `"${Text}"`;
type SingleQuotedStringLiteral<Text extends string = string> = `'${Text}'`;
declare const parseEnclosed: (s: DynamicState, enclosing: EnclosingStartToken) => undefined;
type parseEnclosed<s extends StaticState, enclosingStart extends EnclosingStartToken, unscanned extends string> = Scanner.shiftUntil<unscanned, EnclosingTokens[enclosingStart]> extends Scanner.shiftResult<infer scanned, infer nextUnscanned> ? nextUnscanned extends "" ? state.error<writeUnterminatedEnclosedMessage<scanned, enclosingStart>> : state.setRoot<s, `${enclosingStart}${scanned}${EnclosingTokens[enclosingStart]}`, nextUnscanned extends Scanner.shift<string, infer unscanned> ? unscanned : ""> : never;
declare const enclosingQuote: {
    readonly "'": 1;
    readonly '"': 1;
};
type EnclosingQuote = keyof typeof enclosingQuote;
declare const enclosingTokens: {
    readonly "d'": "'";
    readonly 'd"': "\"";
    readonly "'": "'";
    readonly '"': "\"";
    readonly "/": "/";
};
type EnclosingTokens = typeof enclosingTokens;
type EnclosingStartToken = keyof EnclosingTokens;
type EnclosingEndToken = EnclosingTokens[keyof EnclosingTokens];
declare const enclosingCharDescriptions: {
    readonly '"': "double-quote";
    readonly "'": "single-quote";
    readonly "/": "forward slash";
};
type enclosingCharDescriptions = typeof enclosingCharDescriptions;
declare const writeUnterminatedEnclosedMessage: <fragment extends string, enclosingStart extends "/" | "'" | "\"" | "d'" | "d\"">(fragment: fragment, enclosingStart: enclosingStart) => `${enclosingStart}${fragment} requires a closing ${{
    readonly '"': "double-quote";
    readonly "'": "single-quote";
    readonly "/": "forward slash";
}[{
    readonly "d'": "'";
    readonly 'd"': "\"";
    readonly "'": "'";
    readonly '"': "\"";
    readonly "/": "/";
}[enclosingStart]]}`;
type writeUnterminatedEnclosedMessage<fragment extends string, enclosingStart extends EnclosingStartToken> = `${enclosingStart}${fragment} requires a closing ${enclosingCharDescriptions[EnclosingTokens[enclosingStart]]}`;

type parsedKey<result extends KeyParseResult> = result;
type OptionalTuple<value> = readonly [value, "?"];
type KeyParseResult<kind extends ParsedKeyKind = ParsedKeyKind> = {
    kind: kind;
    innerKey: string | symbol;
};
type EntryParseResult<kind extends ParsedKeyKind = ParsedKeyKind> = extend<KeyParseResult<kind>, {
    innerValue: unknown;
}>;
type ParsedKeyKind = "required" | "optional" | "indexed";
type parsedEntry<result extends EntryParseResult> = result;
type OptionalValue<value> = OptionalStringDefinition<value & string> | OptionalTuple<value>;
type OptionalStringDefinition<name extends string = string> = `${name}?`;
type IndexedKey<def extends string = string> = `[${def}]`;
type validateObjectValue<def, $, args> = def extends OptionalTuple<infer value> ? readonly [validateObjectValueString<value, $, args>, "?"] : validateObjectValueString<def, $, args>;
type validateObjectValueString<def, $, args> = def extends OptionalStringDefinition<infer innerValue> ? `${validateDefinition<innerValue, $, args>}?` : validateDefinition<def, $, args>;
type DefinitionEntry = readonly [string | symbol, unknown];
declare const parseEntry: ([key, value]: DefinitionEntry) => EntryParseResult;
type parseEntry<keyDef extends PropertyKey, valueDef> = parseKey<keyDef> extends infer keyParseResult extends KeyParseResult ? valueDef extends OptionalValue<infer innerValue> ? parsedEntry<{
    kind: "optional";
    innerKey: keyParseResult["innerKey"];
    innerValue: innerValue extends OptionalStringDefinition<infer innerStringValue> ? innerStringValue : innerValue;
}> : parsedEntry<{
    kind: keyParseResult["kind"];
    innerKey: keyParseResult["innerKey"];
    innerValue: valueDef;
}> : never;
type parseKey<k> = k extends OptionalStringDefinition<infer inner> ? inner extends `${infer baseName}${Scanner.EscapeToken}` ? parsedKey<{
    kind: "required";
    innerKey: OptionalStringDefinition<baseName>;
}> : parsedKey<{
    kind: "optional";
    innerKey: inner;
}> : k extends IndexedKey<infer def> ? parsedKey<{
    kind: "indexed";
    innerKey: def;
}> : k extends `${Scanner.EscapeToken}${infer escapedIndexKey extends IndexedKey}` ? parsedKey<{
    kind: "required";
    innerKey: escapedIndexKey;
}> : parsedKey<{
    kind: "required";
    innerKey: k & (string | symbol);
}>;

declare const parseOperand: (s: DynamicState) => void;
type parseOperand<s extends StaticState, $, args> = s["unscanned"] extends Scanner.shift<infer lookahead, infer unscanned> ? lookahead extends "(" ? state.reduceGroupOpen<s, unscanned> : lookahead extends EnclosingEndToken ? parseEnclosed<s, lookahead, unscanned> : lookahead extends Scanner.WhiteSpaceToken ? parseOperand<state.scanTo<s, unscanned>, $, args> : lookahead extends "d" ? unscanned extends Scanner.shift<infer enclosing extends EnclosingQuote, infer nextUnscanned> ? parseEnclosed<s, `d${enclosing}`, nextUnscanned> : parseUnenclosed<s, $, args> : parseUnenclosed<s, $, args> : state.completion<`${s["scanned"]}${BaseCompletions<$, args>}`>;

declare const parseDivisor: (s: DynamicStateWithRoot) => void;
type parseDivisor<s extends StaticState, unscanned extends string> = Scanner.shiftUntilNextTerminator<Scanner.skipWhitespace<unscanned>> extends Scanner.shiftResult<infer scanned, infer nextUnscanned> ? scanned extends NumberLiteral<infer divisor> ? divisor extends 0 ? state.error<writeInvalidDivisorMessage<0>> : state.setRoot<s, [s["root"], "%", scanned], nextUnscanned> : state.error<writeInvalidDivisorMessage<scanned>> : never;
declare const writeInvalidDivisorMessage: <divisor extends string | number>(divisor: divisor) => `% operator must be followed by a non-zero integer literal (was ${divisor})`;
type writeInvalidDivisorMessage<divisor extends string | number> = `% operator must be followed by a non-zero integer literal (was ${divisor})`;

declare const parseOperator: (s: DynamicStateWithRoot) => void;
type parseOperator<s extends StaticState, $, args> = s["unscanned"] extends Scanner.shift<infer lookahead, infer unscanned> ? lookahead extends "[" ? unscanned extends Scanner.shift<"]", infer nextUnscanned> ? state.setRoot<s, [s["root"], "[]"], nextUnscanned> : state.error<incompleteArrayTokenMessage> : lookahead extends "|" | "&" ? state.reduceBranch<s, lookahead, unscanned> : lookahead extends ")" ? state.finalizeGroup<s, unscanned> : Scanner.lookaheadIsFinalizing<lookahead, unscanned> extends true ? state.finalize<state.scanTo<s, unscanned>, lookahead & Scanner.FinalizingLookahead> : lookahead extends ComparatorStartChar ? parseBound<s, lookahead, unscanned, $, args> : lookahead extends "%" ? parseDivisor<s, unscanned> : lookahead extends Scanner.WhiteSpaceToken ? parseOperator<state.scanTo<s, unscanned>, $, args> : state.error<writeUnexpectedCharacterMessage<lookahead>> : state.finalize<s, "">;
declare const writeUnexpectedCharacterMessage: <char extends string, shouldBe extends string>(char: char, shouldBe?: shouldBe) => `'${char}' is not allowed here${shouldBe extends "" ? "" : ` (should be ${shouldBe})`}`;
type writeUnexpectedCharacterMessage<char extends string, shouldBe extends string = ""> = `'${char}' is not allowed here${shouldBe extends "" ? "" : ` (should be ${shouldBe})`}`;
declare const incompleteArrayTokenMessage = "Missing expected ']'";
type incompleteArrayTokenMessage = typeof incompleteArrayTokenMessage;

declare const parseString: (def: string, ctx: ParseContext) => Root;
/**
 * Try to parse the definition from right to left using the most common syntax.
 * This can be much more efficient for simple definitions.
 */
type parseString<def extends string, $, args> = def extends keyof $ ? def : def extends `${infer child}[]` ? child extends keyof $ ? [child, "[]"] : fullStringParse<def, $, args> : fullStringParse<def, $, args>;
type inferString<def extends string, $, args> = inferAst<parseString<def, $, args>, $, args>;
type BaseCompletions<$, args, otherSuggestions extends string = never> = (keyof $ & string) | (keyof args & string) | StringifiablePrefixOperator | otherSuggestions;
declare const fullStringParse: (def: string, ctx: ParseContext) => _arktype_schema.RootNode<"domain", unknown> | _arktype_schema.RootNode<"proto", unknown> | _arktype_schema.RootNode<"unit", unknown> | _arktype_schema.RootNode<"morph", unknown> | _arktype_schema.RootNode<"intersection", unknown> | _arktype_schema.RootNode<"union", unknown>;
type fullStringParse<def extends string, $, args> = extractFinalizedResult<parseUntilFinalizer<state.initialize<def>, $, args>>;
declare const parseUntilFinalizer: (s: DynamicState) => DynamicStateWithRoot;
type parseUntilFinalizer<s extends StaticState, $, args> = s["finalizer"] extends undefined ? parseUntilFinalizer<next<s, $, args>, $, args> : s;
declare const next: (s: DynamicState) => void;
type next<s extends StaticState, $, args> = s["root"] extends undefined ? parseOperand<s, $, args> : parseOperator<s, $, args>;
type extractFinalizedResult<s extends StaticState> = s["finalizer"] extends ErrorMessage ? s["finalizer"] : s["finalizer"] extends "" ? s["root"] : state.error<writeUnexpectedCharacterMessage<`${s["finalizer"]}`>>;

declare const parseUnenclosed: (s: DynamicState) => void;
type parseUnenclosed<s extends StaticState, $, args> = Scanner.shiftUntilNextTerminator<s["unscanned"]> extends Scanner.shiftResult<infer token, infer unscanned> ? token extends "keyof" ? state.addPrefix<s, "keyof", unscanned> : tryResolve<s, token, $, args> extends infer result ? result extends ErrorMessage<infer message> ? state.error<message> : result extends keyof $ ? $[result] extends GenericProps ? parseGenericInstantiation<token, $[result], state.scanTo<s, unscanned>, $, args> : state.setRoot<s, result, unscanned> : state.setRoot<s, result, unscanned> : never : never;
declare const parseGenericInstantiation: (name: string, g: Generic, s: DynamicState) => Root<never>;
type parseGenericInstantiation<name extends string, g extends GenericProps, s extends StaticState, $, args> = Scanner.skipWhitespace<s["unscanned"]> extends `<${infer unscanned}` ? parseGenericArgs<name, g["parameters"], unscanned, $, args> extends infer result ? result extends ParsedArgs<infer argAsts, infer nextUnscanned> ? state.setRoot<s, GenericInstantiationAst<g, argAsts>, nextUnscanned> : result : never : state.error<writeInvalidGenericArgsMessage<name, g["parameters"], []>>;
type tryResolve<s extends StaticState, token extends string, $, args> = token extends keyof $ ? token : token extends keyof args ? token : token extends NumberLiteral ? token : token extends BigintLiteral ? token : token extends `${infer submodule extends keyof $ & string}.${infer reference}` ? $[submodule] extends Module<infer r> ? reference extends keyof r["exports"] ? token : unknown extends r["exports"] ? ErrorMessage<writeNonSubmoduleDotMessage<submodule>> : unresolvableError<s, reference, $[submodule], args, [
    submodule
]> : ErrorMessage<writeNonSubmoduleDotMessage<submodule>> : unresolvableError<s, token, $, args, []>;
declare const writeNonSubmoduleDotMessage: <name extends string>(name: name) => `'${name}' must reference a scope to be accessed using dot syntax`;
type writeNonSubmoduleDotMessage<name extends string> = `'${name}' must reference a scope to be accessed using dot syntax`;
declare const writeMissingSubmoduleAccessMessage: <name extends string>(name: name) => `Reference to submodule '${name}' must specify an alias`;
type writeMissingSubmoduleAccessMessage<name extends string> = `Reference to submodule '${name}' must specify an alias`;
/** Provide valid completions for the current token, or fallback to an
 * unresolvable error if there are none */
type unresolvableError<s extends StaticState, token extends string, $, args, submodulePath extends string[]> = validReferenceFromToken<token, $, args, submodulePath> extends never ? ErrorMessage<writeUnresolvableMessage<qualifiedReference<token, submodulePath>>> : Completion<`${s["scanned"]}${qualifiedReference<validReferenceFromToken<token, $, args, submodulePath>, submodulePath>}`>;
type qualifiedReference<reference extends string, submodulePath extends string[]> = join<[...submodulePath, reference], ".">;
type validReferenceFromToken<token extends string, $, args, submodulePath extends string[]> = Extract<submodulePath extends [] ? BaseCompletions<$, args> : keyof $, `${token}${string}`>;
declare const writeUnresolvableMessage: <token extends string>(token: token) => `'${token}' is unresolvable`;
type writeUnresolvableMessage<token extends string> = `'${token}' is unresolvable`;
type writeMissingRightOperandMessage<token extends string, unscanned extends string = ""> = `Token '${token}' requires a right operand${unscanned extends "" ? "" : ` before '${unscanned}'`}`;
declare const writeMissingRightOperandMessage: <token extends string, unscanned extends string>(token: token, unscanned?: unscanned) => `Token '${token}' requires a right operand${unscanned extends "" ? "" : ` before '${unscanned}'`}`;

type InfixExpression$1 = readonly [unknown, InfixOperator, ...unknown[]];
type validateTuple<def extends List, $, args> = def extends IndexZeroExpression ? validatePrefixExpression<def, $, args> : def extends PostfixExpression ? validatePostfixExpression<def, $, args> : def extends InfixExpression$1 ? validateInfixExpression<def, $, args> : def extends readonly ["", ...unknown[]] | readonly [unknown, "", ...unknown[]] ? readonly [
    def[0] extends "" ? BaseCompletions<$, args, IndexZeroOperator> : def[0],
    def[1] extends "" ? BaseCompletions<$, args, IndexOneOperator> : def[1]
] : validateTupleLiteral<def, $, args>;
type validateTupleLiteral<def extends List, $, args, result extends unknown[] = []> = def extends readonly [infer head, ...infer tail] ? validateTupleLiteral<tail, $, args, [
    ...result,
    validateTupleElement<head, tail, $, args>
]> : result;
type validateTupleElement<head, tail, $, args> = head extends variadicExpression<infer operand> ? validateDefinition<operand, $, args> extends infer syntacticResult ? syntacticResult extends operand ? semanticallyValidateRestElement<operand, $, args> extends infer semanticResult ? semanticResult extends operand ? tail extends [] ? head : prematureRestMessage : semanticResult : never : syntacticResult : never : validateObjectValue<head, $, args>;
type semanticallyValidateRestElement<operand, $, args> = inferDefinition<operand, $, args> extends infer result ? result extends never ? writeNonArrayRestMessage<operand> : isAny<result> extends true ? writeNonArrayRestMessage<operand> : result extends readonly unknown[] ? operand : writeNonArrayRestMessage<operand> : never;
declare const writeNonArrayRestMessage: <operand>(operand: operand) => `Rest element ${operand extends string ? `'${operand}'` : ""} must be an array`;
type writeNonArrayRestMessage<operand> = `Rest element ${operand extends string ? `'${operand}'` : ""} must be an array`;
declare const prematureRestMessage = "Rest elements are only allowed at the end of a tuple";
type prematureRestMessage = typeof prematureRestMessage;
type inferTupleLiteral<def extends List, $, args, result extends unknown[] = []> = def extends readonly [infer head, ...infer tail] ? parseEntry<result["length"], head extends variadicExpression<infer operand> ? operand : head> extends infer entryParseResult extends EntryParseResult ? inferDefinition<entryParseResult["innerValue"], $, args> extends infer element ? head extends variadicExpression ? element extends readonly unknown[] ? inferTupleLiteral<tail, $, args, [...result, ...element]> : never : inferTupleLiteral<tail, $, args, entryParseResult["kind"] extends "optional" ? [...result, element?] : [...result, element]> : never : never : result;
type variadicExpression<operandDef = unknown> = variadicStringExpression<operandDef & string> | variadicTupleExpression<operandDef>;
type variadicStringExpression<operandDef extends string = string> = `...${operandDef}`;
type variadicTupleExpression<operandDef = unknown> = readonly [
    "...",
    operandDef
];
type inferTuple<def extends List, $, args> = def extends TupleExpression ? inferTupleExpression<def, $, args> : inferTupleLiteral<def, $, args>;
type inferTupleExpression<def extends TupleExpression, $, args> = def[1] extends "[]" ? inferDefinition<def[0], $, args>[] : def[1] extends "&" ? inferIntersection<inferDefinition<def[0], $, args>, inferDefinition<def[2], $, args>> : def[1] extends "|" ? inferDefinition<def[0], $, args> | inferDefinition<def[2], $, args> : def[1] extends ":" ? inferNarrow<inferDefinition<def[0], $, args>, def[2]> : def[1] extends "=>" ? parseMorph<def[0], def[2], $, args> : def[1] extends "@" ? inferDefinition<def[0], $, args> : def extends readonly ["===", ...infer values] ? values[number] : def extends readonly [
    "instanceof",
    ...infer constructors extends Constructor[]
] ? InstanceType<constructors[number]> : def[0] extends "keyof" ? inferKeyOfExpression<def[1], $, args> : never;
type validatePrefixExpression<def extends IndexZeroExpression, $, args> = def["length"] extends 1 ? readonly [writeMissingRightOperandMessage<def[0]>] : def[0] extends "keyof" ? readonly [def[0], validateDefinition<def[1], $, args>] : def[0] extends "===" ? readonly [def[0], ...unknown[]] : def[0] extends "instanceof" ? readonly [def[0], ...Constructor[]] : never;
type validatePostfixExpression<def extends PostfixExpression, $, args> = conform<def, readonly [validateDefinition<def[0], $, args>, "[]"]>;
type validateInfixExpression<def extends InfixExpression$1, $, args> = def["length"] extends 2 ? readonly [def[0], writeMissingRightOperandMessage<def[1]>] : readonly [
    validateDefinition<def[0], $, args>,
    def[1],
    def[1] extends "|" ? validateDefinition<def[2], $, args> : def[1] extends "&" ? validateDefinition<def[2], $, args> : def[1] extends ":" ? Predicate<extractIn<inferDefinition<def[0], $, args>>> : def[1] extends "=>" ? Morph<extractOut<inferDefinition<def[0], $, args>>, unknown> : def[1] extends "@" ? BaseAttributes | string : validateDefinition<def[2], $, args>
];
type inferKeyOfExpression<operandDef, $, args> = evaluate<keyof inferDefinition<operandDef, $, args>>;
type TupleExpression = IndexZeroExpression | IndexOneExpression;
type IndexOneOperator = TuplePostfixOperator | TupleInfixOperator;
type TuplePostfixOperator = "[]";
type TupleInfixOperator = "&" | "|" | "=>" | ":" | "@";
type IndexOneExpression<token extends IndexOneOperator = IndexOneOperator> = readonly [unknown, token, ...unknown[]];
type parseMorph<inDef, morph, $, args> = morph extends Morph ? (In: extractIn<inferDefinition<inDef, $, args>>) => Out<inferMorphOut<ReturnType<morph>>> : never;
type MorphAst<i = any, o = any> = (In: i) => Out<o>;
type IndexZeroOperator = "keyof" | "instanceof" | "===";
type IndexZeroExpression<token extends IndexZeroOperator = IndexZeroOperator> = readonly [token, ...unknown[]];

type inferIntersection<l, r> = [l] extends [never] ? never : [r] extends [never] ? never : [l & r] extends [never] ? never : isAny<l | r> extends true ? any : l extends MorphAst<infer lIn, infer lOut> ? r extends MorphAst ? never : (In: evaluate<lIn & r>) => Out<lOut> : r extends MorphAst<infer rIn, infer rOut> ? (In: evaluate<rIn & l>) => Out<rOut> : [l, r] extends [object, object] ? intersectObjects<l, r> extends infer result ? result : never : l & r;
declare class MorphableIntersection extends Hkt.Kind {
    f: (In: conform<this[Hkt.key], [l: unknown, r: unknown]>) => inferIntersection<(typeof In)[0], (typeof In)[1]>;
}
type intersectObjects<l, r> = [l, r] extends [
    infer lList extends List,
    infer rList extends List
] ? intersectArrays<lList, rList, MorphableIntersection> : evaluate<{
    [k in keyof l]: k extends keyof r ? inferIntersection<l[k], r[k]> : l[k];
} & Omit<r, keyof l>>;

type inferAst<ast, $, args> = ast extends List ? inferExpression<ast, $, args> : inferTerminal<ast, $, args>;
type GenericInstantiationAst<g extends GenericProps = GenericProps, argAsts extends unknown[] = unknown[]> = [g, "<>", argAsts];
type inferExpression<ast extends List, $, args> = ast extends GenericInstantiationAst ? inferDefinition<ast[0]["definition"], ast[0]["$"] extends UnparsedScope ? $ : ast[0]["$"], {
    [i in keyof ast[0]["parameters"] & `${number}` as ast[0]["parameters"][i]]: inferAst<ast[2][i & keyof ast[2]], $, args>;
}> : ast[1] extends "[]" ? inferAst<ast[0], $, args>[] : ast[1] extends "|" ? inferAst<ast[0], $, args> | inferAst<ast[2], $, args> : ast[1] extends "&" ? inferIntersection<inferAst<ast[0], $, args>, inferAst<ast[2], $, args>> : ast[1] extends Comparator ? ast[0] extends LimitLiteral ? inferAst<ast[2], $, args> : inferAst<ast[0], $, args> : ast[1] extends "%" ? inferAst<ast[0], $, args> : ast[0] extends "keyof" ? keyof inferAst<ast[1], $, args> : never;
type PostfixOperator = "[]";
type PostfixExpression<operator extends PostfixOperator = PostfixOperator, operand = unknown> = readonly [operand, operator];
type InfixOperator = "|" | "&" | Comparator | "%" | ":" | "=>" | "@";
type InfixExpression<operator extends InfixOperator = InfixOperator, l = unknown, r = unknown> = [l, operator, r];
type RegexLiteral<source extends string = string> = `/${source}/`;
type inferTerminal<token, $, args> = token extends keyof args | keyof $ ? resolve<token, $, args> : token extends StringLiteral<infer Text> ? Text : token extends RegexLiteral ? string : token extends DateLiteral ? Date : token extends NumberLiteral<infer value> ? value : token extends BigintLiteral<infer value> ? value : tryInferSubmoduleReference<$, token>;

type astToString<ast> = `'${astToStringRecurse<ast>}'`;
type astToStringRecurse<ast> = ast extends PostfixExpression<infer operator, infer operand> ? operator extends "[]" ? `${groupAst<operand>}[]` : never : ast extends InfixExpression<infer operator, infer l, infer r> ? operator extends "&" | "|" | "%" | Comparator ? `${groupAst<l>}${operator}${groupAst<r>}` : never : ast extends Stringifiable ? `${ast extends bigint ? `${ast}n` : ast}` : "...";
type groupAst<ast> = ast extends List ? ast[1] extends "[]" ? astToStringRecurse<ast> : `(${astToStringRecurse<ast>})` : astToStringRecurse<ast>;

declare const parseBound: (s: DynamicStateWithRoot, start: ComparatorStartChar) => undefined;
type parseBound<s extends StaticState, start extends ComparatorStartChar, unscanned extends string, $, args> = shiftComparator<start, unscanned> extends infer shiftResultOrError ? shiftResultOrError extends Scanner.shiftResult<infer comparator extends Comparator, infer nextUnscanned> ? s["root"] extends `${infer limit extends LimitLiteral}` ? state.reduceLeftBound<s, limit, comparator, nextUnscanned> : parseRightBound<state.scanTo<s, nextUnscanned>, comparator, $, args> : shiftResultOrError : never;
declare const minComparators: {
    readonly ">": true;
    readonly ">=": true;
};
type MinComparator = keyof typeof minComparators;
declare const maxComparators: {
    readonly "<": true;
    readonly "<=": true;
};
type MaxComparator = keyof typeof maxComparators;
declare const comparators: {
    "==": boolean;
    "<": true;
    "<=": true;
    ">": true;
    ">=": true;
};
type Comparator = keyof typeof comparators;
type LimitLiteral = number | DateLiteral;
declare const oneCharComparators: {
    readonly "<": true;
    readonly ">": true;
};
type OneCharComparator = keyof typeof oneCharComparators;
type ComparatorStartChar = Comparator extends `${infer char}${string}` ? char : never;
declare const shiftComparator: (s: DynamicState, start: ComparatorStartChar) => Comparator;
type shiftComparator<start extends ComparatorStartChar, unscanned extends string> = unscanned extends `=${infer nextUnscanned}` ? [`${start}=`, nextUnscanned] : start extends OneCharComparator ? [start, unscanned] : state.error<singleEqualsMessage>;
declare const singleEqualsMessage = "= is not a valid comparator. Use == to check for equality";
type singleEqualsMessage = typeof singleEqualsMessage;
declare const parseRightBound: (s: DynamicStateWithRoot, comparator: Comparator) => undefined;
type parseRightBound<s extends StaticState, comparator extends Comparator, $, args> = parseOperand<s, $, args> extends infer nextState extends StaticState ? nextState["root"] extends `${infer limit extends LimitLiteral}` ? s["branches"]["leftBound"] extends {} ? comparator extends MaxComparator ? state.reduceRange<s, s["branches"]["leftBound"]["limit"], s["branches"]["leftBound"]["comparator"], comparator, limit, nextState["unscanned"]> : state.error<writeUnpairableComparatorMessage<comparator>> : state.reduceSingleBound<s, comparator, limit, nextState["unscanned"]> : state.error<writeInvalidLimitMessage<comparator, astToString<nextState["root"]>, "right">> : never;
declare const writeInvalidLimitMessage: <comparator extends ">" | ">=" | "<" | "<=" | "==", limit extends string | number, boundKind extends BoundKind>(comparator: comparator, limit: limit, boundKind: boundKind) => `Comparator ${comparator} must be ${boundKind extends "left" ? "preceded" : "followed"} by a corresponding literal (was '${limit}')`;
type writeInvalidLimitMessage<comparator extends Comparator, limit extends string | number, boundKind extends BoundKind> = `Comparator ${comparator} must be ${boundKind extends "left" ? "preceded" : "followed"} by a corresponding literal (was '${limit}')`;
type BoundKind = "left" | "right";
declare const invertedComparators: {
    readonly "<": ">";
    readonly ">": "<";
    readonly "<=": ">=";
    readonly ">=": "<=";
    readonly "==": "==";
};
type InvertedComparators = typeof invertedComparators;

type StringifiablePrefixOperator = "keyof";
type OpenLeftBound = {
    limit: LimitLiteral;
    comparator: MinComparator;
};
declare const writeUnmatchedGroupCloseMessage: <unscanned extends string>(unscanned: unscanned) => `Unmatched )${unscanned extends "" ? "" : ` before ${unscanned}`}`;
type writeUnmatchedGroupCloseMessage<unscanned extends string> = `Unmatched )${unscanned extends "" ? "" : ` before ${unscanned}`}`;
declare const writeUnclosedGroupMessage: <missingChar extends string>(missingChar: missingChar) => `Missing ${missingChar}`;
type writeUnclosedGroupMessage<missingChar extends string> = `Missing ${missingChar}`;
declare const writeOpenRangeMessage: <min extends LimitLiteral, comparator extends ">" | ">=">(min: min, comparator: comparator) => `Left bounds are only valid when paired with right bounds (try ...${comparator}${min})`;
type writeOpenRangeMessage<min extends LimitLiteral, comparator extends MinComparator> = `Left bounds are only valid when paired with right bounds (try ...${comparator}${min})`;
type writeUnpairableComparatorMessage<comparator extends Comparator> = `Left-bounded expressions must specify their limits using < or <= (was ${comparator})`;
declare const writeUnpairableComparatorMessage: <comparator extends ">" | ">=" | "<" | "<=" | "==">(comparator: comparator) => `Left-bounded expressions must specify their limits using < or <= (was ${comparator})`;
declare const writeMultipleLeftBoundsMessage: <openLimit extends LimitLiteral, openComparator extends ">" | ">=", limit extends LimitLiteral, comparator extends ">" | ">=">(openLimit: openLimit, openComparator: openComparator, limit: limit, comparator: comparator) => `An expression may have at most one left bound (parsed ${openLimit}${{
    readonly "<": ">";
    readonly ">": "<";
    readonly "<=": ">=";
    readonly ">=": "<=";
    readonly "==": "==";
}[openComparator]}, ${limit}${{
    readonly "<": ">";
    readonly ">": "<";
    readonly "<=": ">=";
    readonly ">=": "<=";
    readonly "==": "==";
}[comparator]})`;
type writeMultipleLeftBoundsMessage<openLimit extends LimitLiteral, openComparator extends MinComparator, limit extends LimitLiteral, comparator extends MinComparator> = `An expression may have at most one left bound (parsed ${openLimit}${InvertedComparators[openComparator]}, ${limit}${InvertedComparators[comparator]})`;

type GenericDeclaration<name extends string = string, params extends string = string> = `${name}<${params}>`;
type GenericParamsParseError<message extends string = string> = [
    nominal<message, "InvalidGenericParameters">
];
declare const parseGenericParams: (def: string) => string[];
type parseGenericParams<def extends string> = parseParamsRecurse<def, "", [
]> extends infer result extends string[] ? "" extends result[number] ? GenericParamsParseError<emptyGenericParameterMessage> : result : never;
declare const emptyGenericParameterMessage = "An empty string is not a valid generic parameter name";
type emptyGenericParameterMessage = typeof emptyGenericParameterMessage;
type parseParamsRecurse<unscanned extends string, param extends string, result extends string[]> = unscanned extends `${infer lookahead}${infer nextUnscanned}` ? lookahead extends "," ? parseParamsRecurse<nextUnscanned, "", [...result, param]> : lookahead extends Scanner.WhiteSpaceToken ? param extends "" ? parseParamsRecurse<Scanner.skipWhitespace<nextUnscanned>, "", result> : Scanner.skipWhitespace<nextUnscanned> extends `${infer nextNonWhitespace}${infer rest}` ? nextNonWhitespace extends "," ? parseParamsRecurse<rest, "", [...result, param]> : GenericParamsParseError<writeUnexpectedCharacterMessage<nextNonWhitespace, ",">> : [
    ...result,
    param
] : parseParamsRecurse<nextUnscanned, `${param}${lookahead}`, result> : param extends "" ? result : [...result, param];
type ParsedArgs<result extends unknown[] = unknown[], unscanned extends string = string> = {
    result: result;
    unscanned: unscanned;
};
declare const parseGenericArgs: (name: string, params: string[], unscanned: string, ctx: ParseContext) => ParsedArgs<Root[], string>;
type parseGenericArgs<name extends string, params extends string[], unscanned extends string, $, args> = parseGenericArgsRecurse<name, params, unscanned, $, args, [], []>;
declare const parseGenericArgsRecurse: (name: string, params: string[], unscanned: string, ctx: ParseContext, argDefs: string[], argNodes: Root[]) => ParsedArgs<Root[]>;
type parseGenericArgsRecurse<name extends string, params extends string[], unscanned extends string, $, args, argDefs extends string[], argAsts extends unknown[]> = parseUntilFinalizer<state.initialize<unscanned>, $, args> extends infer finalArgState extends StaticState ? {
    defs: [
        ...argDefs,
        finalArgState["scanned"] extends `${infer def}${"," | ">"}` ? def : finalArgState["scanned"]
    ];
    asts: [...argAsts, finalArgState["root"]];
    unscanned: finalArgState["unscanned"];
} extends {
    defs: infer nextDefs extends string[];
    asts: infer nextAsts extends unknown[];
    unscanned: infer nextUnscanned extends string;
} ? finalArgState["finalizer"] extends ">" ? nextAsts["length"] extends params["length"] ? ParsedArgs<nextAsts, nextUnscanned> : state.error<writeInvalidGenericArgsMessage<name, params, nextDefs>> : finalArgState["finalizer"] extends "," ? parseGenericArgsRecurse<name, params, nextUnscanned, $, args, nextDefs, nextAsts> : finalArgState["finalizer"] extends ErrorMessage ? finalArgState : state.error<writeUnclosedGroupMessage<">">> : never : never;
declare const writeInvalidGenericArgsMessage: <name extends string, params extends string[], argDefs extends string[]>(name: name, params: params, argDefs: argDefs) => string;
type writeInvalidGenericArgsMessage<name extends string, params extends string[], argDefs extends string[]> = `${name}<${join<params, ", ">}> requires exactly ${params["length"]} args (got ${argDefs["length"]}${argDefs["length"] extends 0 ? "" : `: ${join<argDefs, ",">}`})`;

type validateRange<l, comparator extends Comparator, r, $, args> = l extends LimitLiteral ? validateBound<r, InvertedComparators[comparator], l, "left", $, args> : l extends [infer leftAst, Comparator, unknown] ? ErrorMessage<writeDoubleRightBoundMessage<astToString<leftAst>>> : validateBound<l, comparator, r & LimitLiteral, "right", $, args>;
type validateBound<boundedAst, comparator extends Comparator, limit extends LimitLiteral, boundKind extends BoundKind, $, args> = inferAst<boundedAst, $, args> extends infer bounded ? [bounded] extends [NumericallyBoundable] ? limit extends number ? validateAst<boundedAst, $, args> : ErrorMessage<writeInvalidLimitMessage<comparator, limit, boundKind>> : bounded extends Date ? limit extends DateLiteral ? validateAst<boundedAst, $, args> : ErrorMessage<writeInvalidLimitMessage<comparator, limit, boundKind>> : ErrorMessage<writeUnboundableMessage<astToString<boundKind extends "left" ? boundedAst[0 & keyof boundedAst] : boundedAst>>> : never;
declare const writeDoubleRightBoundMessage: <root extends string>(root: root) => `Expression ${root} must have at most one right bound`;
type writeDoubleRightBoundMessage<root extends string> = `Expression ${root} must have at most one right bound`;

type validateDivisor<l, $, args> = isDivisible<inferAst<l, $, args>> extends true ? validateAst<l, $, args> : ErrorMessage<writeIndivisibleMessage<astToString<l>>>;
type isDivisible<data> = [data] extends [number] ? true : false;

type validateAst<ast, $, args> = ast extends string ? validateStringAst<ast, $> : ast extends PostfixExpression<infer operator, infer operand> ? operator extends "[]" ? validateAst<operand, $, args> : never : ast extends InfixExpression<infer operator, infer l, infer r> ? operator extends "&" | "|" ? validateInfix<ast, $, args> : operator extends Comparator ? validateRange<l, operator, r, $, args> : operator extends "%" ? validateDivisor<l, $, args> : undefined : ast extends readonly ["keyof", infer operand] ? validateAst<operand, $, args> : ast extends GenericInstantiationAst ? validateGenericArgs<ast["2"], $, args> : ErrorMessage<writeUnexpectedExpressionMessage<astToString<ast>>>;
type writeUnexpectedExpressionMessage<expression extends string> = `Unexpectedly failed to parse the expression resulting from ${expression}`;
type validateGenericArgs<argAsts extends unknown[], $, args> = argAsts extends [
    infer head,
    ...infer tail
] ? validateAst<head, $, args> extends ErrorMessage<infer message> ? ErrorMessage<message> : validateGenericArgs<tail, $, args> : undefined;
type validateStringAst<def extends string, $> = def extends NumberLiteral<infer value> ? number extends value ? ErrorMessage<writeMalformedNumericLiteralMessage<def, "number">> : undefined : def extends BigintLiteral<infer value> ? bigint extends value ? ErrorMessage<writeMalformedNumericLiteralMessage<def, "bigint">> : undefined : def extends keyof $ ? $[def] extends GenericProps ? ErrorMessage<writeInvalidGenericArgsMessage<def, $[def]["parameters"], []>> : $[def] extends Module ? ErrorMessage<writeMissingSubmoduleAccessMessage<def>> : undefined : def extends ErrorMessage ? def : undefined;
type validateString<def extends string, $, args> = validateAst<parseString<def, $, args>, $, args> extends infer result extends ErrorMessage ? result extends Completion<infer text> ? text : result : def;
type validateInfix<ast extends InfixExpression, $, args> = validateAst<ast[0], $, args> extends ErrorMessage<infer message> ? ErrorMessage<message> : validateAst<ast[2], $, args> extends ErrorMessage<infer message> ? ErrorMessage<message> : undefined;

type inferObjectLiteral<def extends object, $, args> = evaluate<{
    -readonly [k in keyof def as nonOptionalKeyFrom<k, def[k], $, args>]: inferDefinition<def[k], $, args>;
} & {
    -readonly [k in keyof def as optionalKeyFrom<k, def[k]>]?: inferDefinition<def[k] extends OptionalValue<infer inner> ? inner : def[k], $, args>;
}>;
type validateObjectLiteral<def, $, args> = {
    [k in keyof def]: k extends IndexedKey<infer indexDef> ? validateString<indexDef, $, args> extends ErrorMessage<infer message> ? indexParseError<message> : inferDefinition<indexDef, $, args> extends PropertyKey ? validateDefinition<def[k], $, args> : indexParseError<writeInvalidPropertyKeyMessage<indexDef>> : validateObjectValue<def[k], $, args>;
};
type nonOptionalKeyFrom<k extends PropertyKey, valueDef, $, args> = parseEntry<k, valueDef> extends infer result extends EntryParseResult ? result["kind"] extends "required" ? result["innerKey"] : result["kind"] extends "indexed" ? inferDefinition<result["innerKey"], $, args> & PropertyKey : never : never;
type optionalKeyFrom<k extends PropertyKey, valueDef> = parseEntry<k, valueDef> extends infer result extends EntryParseResult<"optional"> ? result["innerKey"] : never;
declare const indexParseSymbol: unique symbol;
type indexParseError<message extends string = string> = {
    [indexParseSymbol]: message;
};
declare const writeInvalidPropertyKeyMessage: <indexDef extends string>(indexDef: indexDef) => `Indexed key definition '${indexDef}' must be a string, number or symbol`;
type writeInvalidPropertyKeyMessage<indexDef extends string> = `Indexed key definition '${indexDef}' must be a string, number or symbol`;

type inferDefinition<def, $, args> = isAny<def> extends true ? never : def extends type.cast<infer t> | ThunkCast<infer t> ? t : def extends string ? inferString<def, $, args> : def extends readonly unknown[] ? inferTuple<def, $, args> : def extends RegExp ? string : def extends object ? inferObjectLiteral<def, $, args> : never;
type validateDefinition<def, $, args> = null extends undefined ? ErrorMessage<`'strict' or 'strictNullChecks' must be set to true in your tsconfig's 'compilerOptions'`> : [def] extends [Terminal] ? unknown extends def ? never : def : def extends string ? validateString<def, $, args> : def extends readonly unknown[] ? validateTuple<def, $, args> : def extends BadDefinitionType ? writeBadDefinitionTypeMessage<objectKindOrDomainOf<def>> : isUnknown<def> extends true ? // this allows the initial list of autocompletions to be populated when a user writes "type()",
BaseCompletions<$, args> | {} : validateObjectLiteral<def, $, args>;
type validateDeclared<declared, def, $, args> = def extends validateDefinition<def, $, args> ? validateInference<def, declared, $, args> : validateDefinition<def, $, args>;
type validateInference<def, declared, $, args> = def extends RegExp | type.cast<unknown> | ThunkCast | TupleExpression ? validateShallowInference<def, declared, $, args> : def extends readonly unknown[] ? declared extends readonly unknown[] ? {
    [i in keyof declared]: i extends keyof def ? validateInference<def[i], declared[i], $, args> : unknown;
} : evaluate<declarationMismatch<def, declared, $, args>> : def extends object ? evaluate<{
    [k in requiredKeyOf<declared>]: k extends keyof def ? validateInference<def[k], declared[k], $, args> : unknown;
} & {
    [k in optionalKeyOf<declared> & string as `${k}?`]: `${k}?` extends keyof def ? validateInference<def[`${k}?`], defined<declared[k]>, $, args> : unknown;
}> : validateShallowInference<def, declared, $, args>;
type validateShallowInference<def, declared, $, args> = equals<inferDefinition<def, $, args>, declared> extends true ? def : evaluate<declarationMismatch<def, declared, $, args>>;
type declarationMismatch<def, declared, $, args> = {
    declared: declared;
    inferred: inferDefinition<def, $, args>;
};
type Terminal = RegExp | type.cast<unknown> | ((...args: never[]) => unknown);
type ThunkCast<t = unknown> = () => type.cast<t>;
type BadDefinitionType = Exclude<Primitive, string>;
declare const writeBadDefinitionTypeMessage: <actual extends string>(actual: actual) => `Type definitions must be strings or objects (was ${actual})`;
type writeBadDefinitionTypeMessage<actual extends string> = `Type definitions must be strings or objects (was ${actual})`;

type TypeParser<$> = {
    <const def>(def: validateTypeRoot<def, $>): Type<inferTypeRoot<def, $>, $>;
    <const zero, const one, const rest extends readonly unknown[]>(_0: zero extends IndexZeroOperator ? zero : validateTypeRoot<zero, $>, _1: zero extends "keyof" ? validateTypeRoot<one, $> : zero extends "instanceof" ? conform<one, Constructor> : zero extends "===" ? conform<one, unknown> : conform<one, IndexOneOperator>, ..._2: zero extends "===" ? rest : zero extends "instanceof" ? conform<rest, readonly Constructor[]> : one extends TupleInfixOperator ? one extends ":" ? [Predicate<extractIn<inferTypeRoot<zero, $>>>] : one extends "=>" ? [
        Morph<extractOut<inferTypeRoot<zero, $>>, unknown>
    ] : one extends "@" ? [string | BaseAttributes] : [validateTypeRoot<rest[0], $>] : []): Type<inferTypeRoot<[zero, one, ...rest], $>, $>;
    <params extends string, const def>(params: `<${validateParameterString<params>}>`, def: validateDefinition<def, $, {
        [param in parseGenericParams<params>[number]]: unknown;
    }>): Generic<parseGenericParams<params>, def, $>;
};
type DeclarationParser<$> = <preinferred>() => {
    type: <def>(def: validateDeclared<preinferred, def, $, bindThis<def>>) => Type<preinferred, $>;
};
type arkKind = typeof arkKind;
type DefinitionParser<$> = <def>(def: validateDefinition<def, $, bindThis<def>>) => def;
type KeyCheckKind = "distilled" | "strict" | "loose";
type TypeConfig = {
    keys?: KeyCheckKind;
    mustBe?: string;
};
declare class Type<t = unknown, $ = any> extends CompiledFunction<(data: unknown) => CheckResult<extractOut<t>>> {
    definition: unknown;
    scope: Scope;
    [inferred]: t;
    inferMorph: t;
    infer: extractOut<t>;
    inferIn: extractIn<t>;
    config: TypeConfig;
    root: Root<t>;
    condition: string;
    alias: string;
    allows: this["root"]["allows"];
    json: Json;
    constructor(definition: unknown, scope: Scope);
    configure(config: TypeConfig): this;
    from(literal: this["infer"]): this["infer"];
    fromIn(literal: this["inferIn"]): this["inferIn"];
    and<def>(def: validateTypeRoot<def, $>): Type<inferIntersection<t, inferTypeRoot<def, $>>, $>;
    or<def>(def: validateTypeRoot<def, $>): Type<t | inferTypeRoot<def, $>, $>;
    morph<morph extends Morph<extractOut<t>>>(morph: morph): Type<(In: this["inferIn"]) => Out<inferMorphOut<ReturnType<morph>>>, $>;
    morph<morph extends Morph<extractOut<t>>, def>(morph: morph, outValidator: validateTypeRoot<def, $>): Type<(In: this["inferIn"]) => Out<extractOut<inferTypeRoot<def, $>>>, $>;
    narrow<def extends Predicate<extractOut<t>>>(def: def): Type<includesMorphs<t> extends true ? (In: this["inferIn"]) => Out<inferNarrow<this["infer"], def>> : inferNarrow<this["infer"], def>, $>;
    array(): Type<t[], $>;
    keyof(): Type<keyof this["inferIn"], $>;
    assert(data: unknown): extractOut<t>;
    equals<other>(other: Type<other>): this is Type<other, $>;
    extends<other>(other: Type<other>): this is Type<other, $>;
}
type validateTypeRoot<def, $> = validateDefinition<def, $, bindThis<def>>;
type inferTypeRoot<def, $> = inferDefinition<def, $, bindThis<def>>;
type validateParameterString<params extends string> = parseGenericParams<params> extends GenericParamsParseError<infer message> ? message : params;
type GenericProps<params extends string[] = string[], def = unknown, $ = any> = {
    [arkKind]: "generic";
    $: $;
    parameters: params;
    definition: def;
    scope: Scope;
};
type Generic<params extends string[] = string[], def = unknown, $ = any> = {
    <args>(...args: conform<args, {
        [i in keyof params]: validateTypeRoot<args[i & keyof args], $>;
    }>): Type<inferDefinition<def, $, bindGenericInstantiation<params, $, args>>, $>;
} & GenericProps<params, def, $>;
type bindGenericInstantiation<params extends string[], $, args> = {
    [i in keyof params & `${number}` as params[i]]: inferTypeRoot<args[i & keyof args], $>;
};
type extractIn<t> = includesMorphs<t> extends true ? extractMorphs<t, "in"> : t;
type extractOut<t> = includesMorphs<t> extends true ? extractMorphs<t, "out"> : t;
type includesMorphs<t> = [
    t,
    extractMorphs<t, "in">,
    t,
    extractMorphs<t, "out">
] extends [extractMorphs<t, "in">, t, extractMorphs<t, "out">, t] ? false : true;
type extractMorphs<t, io extends "in" | "out"> = t extends MorphAst<infer i, infer o> ? io extends "in" ? i : o : t extends TerminallyInferredObjectKind | Primitive ? t : {
    [k in keyof t]: extractMorphs<t[k], io>;
};
/** Objects we don't want to expand during inference like Date or Promise */
type TerminallyInferredObjectKind = ReturnType<ArkConfig["preserve"]> | BuiltinObjects[Exclude<BuiltinObjectKind, "Object" | "Array">];

type cedille = "¸";
type serializedWhentry<k extends string, v extends string> = `[${k}${cedille}${v}]`;
type parseWhentryKey<s extends string, $, result = {}> = s extends `${serializedWhentry<infer k, infer v>}${cedille}${infer tail}` ? validateTypeRoot<v, $> extends ErrorMessage<infer message> ? ErrorMessage<message> : parseWhentryKey<tail, $, result & {
    [_ in k]: inferTypeRoot<v, $>;
}> : s extends serializedWhentry<infer k, infer v> ? validateTypeRoot<v, $> extends ErrorMessage<infer message> ? ErrorMessage<message> : evaluate<result & {
    [_ in k]: inferTypeRoot<v, $>;
}> : validateTypeRoot<s, $> extends ErrorMessage<infer message> ? ErrorMessage<message> : never;
type WhenParser<$> = <const def>(def: validateTypeRoot<def, $>) => join<unionToTuple<`[${join<conform<entryOf<def>, [string, string]>, cedille>}]`>, cedille>;
type MatchContext = {
    inConstraint: unknown;
    outConstraint: unknown;
    thens: readonly Fn[];
    $: unknown;
};
type validateCases<cases, ctx extends MatchContext> = {
    [k in keyof cases | keyof ctx["$"]]?: k extends validateTypeRoot<k, ctx["$"]> ? (In: ctx["inConstraint"] & inferTypeRoot<k, ctx["$"]>) => ctx["outConstraint"] : parseWhentryKey<k & string, ctx["$"]> extends ErrorMessage<infer message> ? ErrorMessage<message> : (In: parseWhentryKey<k & string, ctx["$"]>) => ctx["outConstraint"];
};
type MatchParser<$> = {
    <In = unknown, Out = unknown>(): ChainableMatchParser<{
        inConstraint: In;
        outConstraint: Out;
        thens: [];
        $: $;
    }>;
} & CaseMatchParser<{
    inConstraint: unknown;
    outConstraint: unknown;
    thens: [];
    $: $;
}>;
type CaseMatchParser<ctx extends MatchContext> = <cases>(def: conform<cases, validateCases<cases, ctx>>) => ChainableMatchParser<replaceKey<ctx, "thens", [...ctx["thens"], ...unionToTuple<valueOf<cases>>]>>;
type WhenMatchParser<ctx extends MatchContext> = <def, then extends (In: ctx["inConstraint"] & inferTypeRoot<def, ctx["$"]>) => ctx["outConstraint"]>(when: validateTypeRoot<def, ctx["$"]>, then: then) => ChainableMatchParser<replaceKey<ctx, "thens", [...ctx["thens"], then]>>;
type MatchInvokation<ctx extends MatchContext> = <data extends ctx["inConstraint"]>(In: data) => {
    [i in Extract<keyof ctx["thens"], `${number}`>]: isDisjoint<data, paramsOf<ctx["thens"][i]>[0]> extends true ? never : returnOf<ctx["thens"][i]>;
}[Extract<keyof ctx["thens"], `${number}`>];
type ChainableMatchParser<ctx extends MatchContext> = MatchInvokation<ctx> & {
    cases: CaseMatchParser<ctx>;
    when: WhenMatchParser<ctx>;
};

type ScopeParser<parent, ambient> = {
    <const def>(def: validateScope<def, parent & ambient>): Scope<{
        exports: inferBootstrapped<{
            exports: bootstrapExports<def>;
            locals: bootstrapLocals<def> & parent;
            ambient: ambient;
        }>;
        locals: inferBootstrapped<{
            exports: bootstrapLocals<def>;
            locals: bootstrapExports<def> & parent;
            ambient: ambient;
        }>;
        ambient: ambient;
    }>;
};
type ScopeConfig = {
    ambient?: Scope | null;
    codes?: Record<ProblemCode, {
        mustBe?: string;
    }>;
    keys?: KeyCheckKind;
};
type validateScope<def, $> = {
    [k in keyof def]: parseScopeKey<k>["params"] extends [] ? def[k] extends Type | PreparsedResolution ? def[k] : validateDefinition<def[k], $ & bootstrap<def>, {}> : parseScopeKey<k>["params"] extends GenericParamsParseError ? parseScopeKey<k>["params"][0] : validateDefinition<def[k], $ & bootstrap<def>, {
        [param in parseScopeKey<k>["params"][number]]: unknown;
    }>;
};
type bindThis<def> = {
    this: Def<def>;
};
declare const bindThis: () => {
    this: _arktype_schema.RootNode<"intersection", unknown>;
};
/** nominal type for an unparsed definition used during scope bootstrapping */
type Def<def = {}> = nominal<def, "unparsed">;
/** sentinel indicating a scope that will be associated with a generic has not yet been parsed */
type UnparsedScope = "$";
type bootstrap<def> = bootstrapLocals<def> & bootstrapExports<def>;
type bootstrapLocals<def> = bootstrapAliases<{
    [k in keyof def & PrivateDeclaration as extractPrivateKey<k>]: def[k];
}>;
type bootstrapExports<def> = bootstrapAliases<{
    [k in Exclude<keyof def, PrivateDeclaration>]: def[k];
}>;
/** These are legal as values of a scope but not as definitions in other contexts */
type PreparsedResolution = Module | GenericProps;
type bootstrapAliases<def> = {
    [k in Exclude<keyof def, GenericDeclaration | symbol>]: def[k] extends PreparsedResolution ? def[k] : def[k] extends (() => infer thunkReturn extends PreparsedResolution) ? thunkReturn : Def<def[k]>;
} & {
    [k in keyof def & GenericDeclaration as extractGenericName<k>]: Generic<parseGenericParams<extractGenericParameters<k>>, def[k], UnparsedScope>;
};
type inferBootstrapped<r extends Resolutions> = evaluate<{
    [name in keyof r["exports"]]: r["exports"][name] extends Def<infer def> ? inferDefinition<def, $<r>, {}> : r["exports"][name] extends GenericProps<infer params, infer def> ? Generic<params, def, $<r>> : r["exports"][name];
}>;
type extractGenericName<k> = k extends GenericDeclaration<infer name> ? name : never;
type extractGenericParameters<k> = k extends GenericDeclaration<string, infer params> ? params : never;
type extractPrivateKey<k> = k extends PrivateDeclaration<infer key> ? key : never;
type PrivateDeclaration<key extends string = string> = `#${key}`;
type resolve<reference extends keyof $ | keyof args, $, args> = (reference extends keyof args ? args[reference] : $[reference & keyof $]) extends infer resolution ? [resolution] extends [never] ? never : isAny<resolution> extends true ? any : resolution extends Def<infer def> ? inferDefinition<def, $, args> : resolution : never;
type moduleKeyOf<$> = {
    [k in keyof $]: $[k] extends Module ? k & string : never;
}[keyof $];
type tryInferSubmoduleReference<$, token> = token extends `${infer submodule extends moduleKeyOf<$>}.${infer subalias}` ? subalias extends keyof $[submodule] ? $[submodule][subalias] extends Type<infer t> ? t : never : never : never;
type $<r extends Resolutions> = r["exports"] & r["locals"] & r["ambient"];
type exportedName<r extends Resolutions> = keyof r["exports"] & string;
type Module<r extends Resolutions = any> = {
    [k in exportedName<r> | arkKind]: k extends string ? [r["exports"][k]] extends [never] ? Type<never, $<r>> : isAny<r["exports"][k]> extends true ? Type<any, $<r>> : r["exports"][k] extends PreparsedResolution ? r["exports"][k] : Type<r["exports"][k], $<r>> : type.cast<"module">;
};
type Resolutions = {
    exports: unknown;
    locals: unknown;
    ambient: unknown;
};
type ParseContext = {
    baseName: string;
    path: string[];
    scope: Scope;
    args: Record<string, Root> | undefined;
};
type ParseContextInput = Pick<ParseContext, "baseName" | "args">;
declare class Scope<r extends Resolutions = any> {
    infer: extractOut<r["exports"]>;
    inferIn: extractIn<r["exports"]>;
    config: TypeConfig;
    private parseCache;
    private resolutions;
    /** The set of names defined at the root-level of the scope mapped to their
     * corresponding definitions.**/
    aliases: Record<string, unknown>;
    private exportedNames;
    private ambient;
    private references;
    constructor(def: Dict, config: ScopeConfig);
    static root: ScopeParser<{}, {}>;
    type: TypeParser<$<r>>;
    match: MatchParser<$<r>>;
    when: WhenParser<$<r>>;
    declare: DeclarationParser<$<r>>;
    scope: ScopeParser<r["exports"], r["ambient"]>;
    define: DefinitionParser<$<r>>;
    toAmbient(): Scope<{
        exports: r["exports"];
        locals: r["locals"];
        ambient: r["exports"];
    }>;
    get<name extends keyof r["exports"] & string>(name: name): Type<r["exports"][name], $<r>>;
    private createRootContext;
    parseDefinition(def: unknown, input: ParseContextInput): Root;
    parse(def: unknown, ctx: ParseContext): Root;
    maybeResolve(name: string): Root | Generic | undefined;
    /** If name is a valid reference to a submodule alias, return its resolution  */
    private maybeResolveSubalias;
    maybeResolveNode(name: string): Root | undefined;
    import<names extends exportedName<r>[]>(...names: names): destructuredImportContext<r, names extends [] ? keyof r["exports"] : names[number]>;
    compile(): string;
    private exportedResolutions;
    private exportCache;
    export<names extends exportedName<r>[]>(...names: names): Module<names extends [] ? r : destructuredExportContext<r, names[number]>>;
}
type destructuredExportContext<r extends Resolutions, name extends exportedName<r>> = {
    exports: {
        [k in name]: r["exports"][k];
    };
    locals: r["locals"] & {
        [k in Exclude<keyof r["exports"], name>]: r["exports"][k];
    };
    ambient: r["ambient"];
};
type destructuredImportContext<r extends Resolutions, name extends exportedName<r>> = {
    [k in name as `#${k & string}`]: type.cast<r["exports"][k]>;
};
type ParsedScopeKey = {
    isLocal: boolean;
    name: string;
    params: string[];
};
declare const parseScopeKey: (k: string) => ParsedScopeKey;
type parseScopeKey<k> = k extends PrivateDeclaration<infer inner> ? parsePossibleGenericDeclaration<inner, true> : parsePossibleGenericDeclaration<k, false>;
type parsePossibleGenericDeclaration<k, isLocal extends boolean> = k extends GenericDeclaration<infer name, infer paramString> ? {
    isLocal: isLocal;
    name: name;
    params: parseGenericParams<paramString>;
} : {
    isLocal: isLocal;
    name: k;
    params: [];
};

export { Ark, Module, Scope, Type, ark, arktypes, declare, define, inferTypeRoot, jsObjects, match, scope, tsGenerics, tsKeywords, type, validateTypeRoot, validation, when };
